# -*- coding=utf-8 -*-

"""Unit testing for assembly graph."""

# pylint: disable=compare-to-zero, missing-raises-doc

from pathlib import Path

from khloraascaf.exceptions import (
    CombineScaffoldingError,
    NotACircuit,
    RepeatScaffoldingError,
    ScaffoldingError,
    UnfeasibleDR,
    UnfeasibleIR,
    UnfeasibleUN,
    UniqueScaffoldingError,
    WrongRegionCode,
    WrongSolverName,
)
from khloraascaf.lib import IR_CODE, UN_CODE


# ============================================================================ #
#                                TEST FUNCTIONS                                #
# ============================================================================ #
def test_scaffolding_error():
    """Test ScaffoldingError exception."""
    exc = ScaffoldingError(Path('./jaaj'))
    assert str(exc) == (
        'Scaffolding fails\n'
        '\tSee output directory: jaaj'
    )


def test_repeat_scaffolding_error():
    """Test RepeatScaffoldingError exception."""
    exc = RepeatScaffoldingError()
    assert str(exc) == 'One repeat scaffolding has failed'


def test_unique_scaffolding_error():
    """Test UniqueScaffoldingError exception."""
    exc = UniqueScaffoldingError()
    assert str(exc) == 'One unique region scaffolding has failed'


def test_combine_scaffolding_error():
    """Test CombineScaffoldingError exception."""
    exc = CombineScaffoldingError()
    assert str(exc) == 'The scaffolding combination has failed'


def test_wrong_region_code():
    """Test WrongRegionCode exception."""
    exc = WrongRegionCode(IR_CODE)
    assert str(exc) == 'The region code 1 is not correct'


def test_wrong_solver_name():
    """Test WrongSolverName exception."""
    exc = WrongSolverName('wrong_solver')
    assert str(exc) == 'The solver name wrong_solver is not correct'


def test_unfeasible_ir():
    """Test UnfeasibleIR exception."""
    exc = UnfeasibleIR(
        -1, (IR_CODE, UN_CODE),
    )
    assert str(exc) == (
        'The Find the best inverted repeats problem is unfeasible:\n'
        '\t* ILP codes: ir-un\n'
        '\t* Status: Infeasible'
    )


def test_unfeasible_dr():
    """Test UnfeasibleDR exception."""
    exc = UnfeasibleDR(
        -1, (IR_CODE, UN_CODE),
    )
    assert str(exc) == (
        'The Find the best direct repeats problem is unfeasible:\n'
        '\t* ILP codes: ir-un\n'
        '\t* Status: Infeasible'
    )


def test_unfeasible_un():
    """Test UnfeasibleUN exception."""
    exc = UnfeasibleUN(
        -1, (IR_CODE, UN_CODE),
    )
    assert str(exc) == (
        'The Find the best unique regions problem is unfeasible:\n'
        '\t* ILP codes: ir-un\n'
        '\t* Status: Infeasible'
    )


def test_not_a_circuit():
    """Test NotACircuit exception."""
    exc = NotACircuit()
    assert str(exc) == 'The found path is not a circuit'



# -*- coding=utf-8 -*-

"""Output of scaffolding subcommand."""

from datetime import datetime
from pathlib import Path
from typing import Iterable, Iterator, Literal, Optional

from revsymg.index_lib import IndexT, IndOrT, OrT

from khloraascaf.inputs import (
    FORWARD_STR,
    REVERSE_STR,
    STR_ORIENT,
    IdCT,
    OrStrT,
)
from khloraascaf.multiplied_doubled_contig_graph import (
    CIND_IND,
    COR_IND,
    MDCGraphIDContainer,
)
from khloraascaf.result import ScaffoldingResult


# DOCU missing docstrings for constants
# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
OrCT = tuple[IdCT, OrT]
"""
Oriented contig type.

:alias of: :class:`tuple` (:class:`IdCT`, :class:`IndexT`)
"""
# DOCU any role cannot find IndexT
# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                 For OrCT Type                                #
# ---------------------------------------------------------------------------- #
ORC_ID_IND: Literal[0] = 0
"""Index of the identifier for :class:`~revsymg.index_lib.OrT`"""
ORC_OR_IND: Literal[1] = 1
"""Index of the orientation for :class:`~revsymg.index_lib.OrT`"""

# ---------------------------------------------------------------------------- #
#                                 Orientations                                 #
# ---------------------------------------------------------------------------- #
ORIENT_INT_STR: tuple[OrStrT, OrStrT] = (FORWARD_STR, REVERSE_STR)
"""Get orientation strings from orientation binaries.

:type: :class:`tuple` (:class:`str`, :class:`str`)
"""

# ---------------------------------------------------------------------------- #
#                                     Files                                    #
# ---------------------------------------------------------------------------- #
#
# Contigs of the Regions
#
CONTIGS_OF_REGIONS_PREFIX = 'contigs_of_regions'
"""Prefix of the contigs of regions file name."""
CONTIGS_OF_REGIONS_EXT = 'tsv'
"""Extension of the contigs of regions file"""
#
# Map of the regions
#
MAP_OF_REGIONS_PREFIX = 'map_of_regions'
"""Prefix of the map of regions file name."""
MAP_OF_REGIONS_EXT = 'tsv'
"""Extension of the map of regions file"""


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                          Default Output Directories                          #
# ---------------------------------------------------------------------------- #
def generate_output_directory(instance_name: str) -> Path:
    """Return an unique output directory path.

    The output directory name respects the following format:
    ``<yyyy>-<mm>-<dd>_<HH>:<MM>:<SS>_<instance_name>``

    Parameters
    ----------
    instance_name : str
        Instance's name

    Returns
    -------
    Path
        Output directory
    """
    # DOCU output directory generated function
    return Path(
        datetime.now().strftime('%Y-%m-%d_%H:%M:%S') + '_' + instance_name,
    )


# ---------------------------------------------------------------------------- #
#                              Contigs Of Regions                              #
# ---------------------------------------------------------------------------- #
def write_contigs_of_regions(
        result: ScaffoldingResult, contigs_of_regions_path: Path,
        id_container: Optional[MDCGraphIDContainer] = None):
    """Write the regions' contigs file.

    #DOCU describe the output file here

    Parameters
    ----------
    result : ScaffoldingResult
        Previous scaffolding result
    contigs_of_regions_path : Path
        Path of the future file containing the regions' oriented contigs
    id_container : MDCGraphIDContainer, optional
        Identifiers container for the graph, by default None
    """
    # TOTEST function needs ScaffoldingResult & MDCGraphIDContainer fixtures
    with open(contigs_of_regions_path, 'w', encoding='utf-8') as f_out:
        for region_index in range(result.number_regions()):
            line: str = ''
            for occorc in result.region_occorc(region_index):
                if id_container is not None:
                    line += str(
                        id_container.vertex_to_contig(occorc[CIND_IND]),
                    ) + '\t'
                else:
                    line += f'\t{occorc[CIND_IND]}'
                line += f'{ORIENT_INT_STR[occorc[COR_IND]]}\t'
            f_out.write(line[:-1] + '\n')


def read_contigs_of_regions(contigs_of_regions_path: Path) -> (
        Iterator[list[OrCT]]):
    """Read the regions' contigs file.

    Parameters
    ----------
    contigs_of_regions_path : Path
        List of oriented contigs for each region

    Yields
    ------
    list of OrCT
        List of oriented contigs for one region
    """
    with open(contigs_of_regions_path, 'r', encoding='utf-8') as cor_in:
        for line in cor_in:
            orc_of_region: list[OrCT] = []
            l_orc = line.split()
            k = 0
            while k < len(l_orc) - 1:
                orc_of_region.append(
                    (l_orc[k], STR_ORIENT[l_orc[k + 1]]),  # type: ignore
                )
                k += 2
            yield orc_of_region


def fmt_contigs_of_regions_filename(instance_name: str,
                                    ilp_combination: Iterable[str]) -> str:
    """Format the contigs of regions filename.

    Parameters
    ----------
    instance_name : str
        Instance name
    ilp_combination : iterable of str
        ILP string code combination

    Returns
    -------
    str
        Formatted filename
    """
    return (
        f'{CONTIGS_OF_REGIONS_PREFIX}_{instance_name}'
        f"_{'_'.join(ilp_combination)}"
        f'.{CONTIGS_OF_REGIONS_EXT}'
    )


# ---------------------------------------------------------------------------- #
#                                Map Of Regions                                #
# ---------------------------------------------------------------------------- #
def write_map_of_regions(result: ScaffoldingResult,
                         map_of_regions_path: Path):
    """Write the region map file.

    # DOCU describe the output file here

    Parameters
    ----------
    result : ScaffoldingResult
        Scaffolding result
    map_of_regions_path : Path
        Path of the future file containing the regions' oriented contigs
    """
    with open(map_of_regions_path, 'w', encoding='utf-8') as f_out:
        for reg_ind, reg_or in result.map_of_regions():
            f_out.write(f'{reg_ind}\t{ORIENT_INT_STR[reg_or]}\n')


def read_map_of_regions(map_of_regions_path: Path) -> Iterator[IndOrT]:
    """Write the region map file.

    Parameters
    ----------
    map_of_regions_path : Path
        Map of regions

    Yields
    ------
    IndOrT
        The index and the orientation of the region
    """
    with open(map_of_regions_path, 'r', encoding='utf-8') as mof_in:
        for line in mof_in:
            reg_indstr, reg_orstr = line.split()
            yield IndexT(reg_indstr), STR_ORIENT[reg_orstr]  # type: ignore


def fmt_map_of_regions_filename(instance_name: str,
                                ilp_combination: Iterable[str]) -> str:
    """Format map of the regions filename.

    Parameters
    ----------
    instance_name : str
        Instance name
    ilp_combination : iterable of str
        ILP string code combination

    Returns
    -------
    str
        Formatted filename
    """
    return (
        f'{MAP_OF_REGIONS_PREFIX}_{instance_name}'
        f"_{'_'.join(ilp_combination)}"
        f'.{MAP_OF_REGIONS_EXT}'
    )



# -*- coding=utf-8 -*-

"""Unit testing for utils_debug module."""
from pathlib import Path

from revsymg.index_lib import FORWARD_INT, REVERSE_INT

from khloraascaf.inputs import INSTANCE_NAME_DEF
from khloraascaf.lib import DR_REGION_ID, IR_REGION_ID, UN_REGION_ID
from khloraascaf.utils_debug import (
    DIRF_CODE,
    fmt_found_repeated_fragments_filename,
    fmt_vertices_of_regions_filename,
    read_found_repeated_fragments,
    read_vertices_of_regions,
)


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
TEST_DIR: Path = Path(__file__).parent.absolute()
_TOY_DATADIR = TEST_DIR / 'data'
_IR_DR_UN_DIR = _TOY_DATADIR / 'ir_dr_un'

_VERTICES_OF_REGION_PATH = (
    _IR_DR_UN_DIR
    / 'vertices_of_regions_khloraascaf_ir_dr_un.tsv'
)

_REPFRAG_PATH = (
    _IR_DR_UN_DIR
    / 'repfrag_khloraascaf_ir_dr.tsv'
)


# ============================================================================ #
#                                TEST FUNCTIONS                                #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                              Vertices Of Regions                             #
# ---------------------------------------------------------------------------- #
def test_read_vertices_of_regions():
    """Test for read_vertices_of_regions."""
    assert tuple(
        read_vertices_of_regions(
            _VERTICES_OF_REGION_PATH,
        ),
    ) == (
        [(0, FORWARD_INT, 0), (1, REVERSE_INT, 0)],
        [(2, FORWARD_INT, 1), (3, REVERSE_INT, 1)],
        [(4, FORWARD_INT, 0)],
        [(5, FORWARD_INT, 0), (6, REVERSE_INT, 1), (7, REVERSE_INT, 1)],
        [(8, REVERSE_INT, 0)],
        [(9, FORWARD_INT, 0)],
    )


def test_fmt_vertices_of_regions_filename():
    """Test for fmt_vertices_of_regions_filename."""
    assert fmt_vertices_of_regions_filename(
        INSTANCE_NAME_DEF, (IR_REGION_ID, DR_REGION_ID, UN_REGION_ID),
    ) == _VERTICES_OF_REGION_PATH.name


# ---------------------------------------------------------------------------- #
#                           Found Repeated Fragments                           #
# ---------------------------------------------------------------------------- #
def test_read_found_repeated_fragments():
    """Test for read_found_repeated_fragments."""
    assert tuple(
        read_found_repeated_fragments(_REPFRAG_PATH),
    ) == (
        (DIRF_CODE, (2, FORWARD_INT, 0)),
        (DIRF_CODE, (3, REVERSE_INT, 0)),
    )


def test_fmt_found_repeated_fragments_filename():
    """Test for fmt_found_repeated_fragments_filename."""
    assert fmt_found_repeated_fragments_filename(
        INSTANCE_NAME_DEF, (IR_REGION_ID, DR_REGION_ID),
    ) == _REPFRAG_PATH.name



# -*- coding=utf-8 -*-

"""Module for assembly graph and multiple solution generator."""


from collections.abc import Iterable, Iterator
from pathlib import Path
from queue import LifoQueue

from bitarray import bitarray
from revsymg.graphs import RevSymGraph
from revsymg.index_lib import (
    FORWARD_INT,
    IND,
    ORIENT_REV,
    IndexT,
    IndOrIndT,
    IndOrT,
    OrT,
)

from khloraascaf.inputs import STR_ORIENT
from khloraascaf.outputs import (
    ORC_ID_IND,
    ORC_OR_IND,
    ORIENT_INT_STR,
    OrCT,
    read_contigs_of_regions,
    read_map_of_regions,
)


# DOCU: all assembly graph
# DOCU: tuto for assembly graph
# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     Files                                    #
# ---------------------------------------------------------------------------- #
#
# Region paths
#
REGION_PATHS_PREFIX = 'region_paths'
"""Prefix of the region paths file name."""
REGION_PATHS_EXT = 'tsv'
"""Extension of the region paths file"""
#
# Oriented contig paths
#
ORIENTED_CONTIG_PATHS_PREFIX = 'oriented_contig_paths'
"""Prefix of the oriented contig paths file name."""
ORIENTED_CONTIG_PATHS_EXT = 'tsv'
"""Extension of the oriented contig paths file"""


# ============================================================================ #
#                                     CLASS                                    #
# ============================================================================ #
class AssemblyGraph():
    """Khloraascaf assembly graph."""

    def __init__(self, map_of_regions_path: Path,
                 contigs_of_regions_path: Path):
        """The Initializer."""
        self.__graph: RevSymGraph = RevSymGraph()
        self.__regions_contigs: list[list[OrCT]] = []
        self.__region_path_length: int = 0
        self.__read_regions_contigs(contigs_of_regions_path)
        self.__add_regions_links(map_of_regions_path)

    def all_region_paths(self) -> Iterator[list[IndOrT]]:  # noqa
        """Iterate over all the paths of oriented regions.

        Yields
        ------
        list of IndOrT
            Path of oriented regions
        """
        edges = self.__graph.edges()

        used_edges: bitarray = bitarray('0') * (
            edges.biggest_edge_index() + 1)

        starter = (0, FORWARD_INT)

        path: list[IndOrT] = [starter]
        eind_path: list[IndexT] = []
        #
        # The lifo contains the number of elements to keep to continue
        #   at the next remaning branch (first IndexT value)
        #
        lifo: LifoQueue[tuple[IndexT, IndOrT, IndexT]] = LifoQueue()

        for v, e_ind in edges.succs(starter):
            if v[IND] == starter[IND]:
                yield path
            else:
                lifo.put((1, v, e_ind))

        while not lifo.empty():
            _, v, e_ind = lifo.get()
            path.append(v)
            eind_path.append(e_ind)
            used_edges[e_ind] = True

            vw_not_used: list[IndOrIndT] = []
            for w, f_ind in edges.succs(v):
                if w[IND] != starter[IND] and not used_edges[f_ind]:
                    vw_not_used.append((w, f_ind))

            if vw_not_used:
                for w, f_ind in vw_not_used:
                    lifo.put((len(path), w, f_ind))
            else:
                if len(path) == self.__region_path_length:
                    yield path
                #
                # Back to the next branching vertex
                #
                if not lifo.empty():
                    branch_src_succ_eind = lifo.get()
                    lifo.put(branch_src_succ_eind)
                    for _ in range(len(path) - branch_src_succ_eind[0]):
                        path.pop()
                        used_edges[eind_path.pop()] = False

    def all_oriented_contig_paths(self) -> Iterator[list[OrCT]]:
        """Iterate over all the paths of oriented contigs.

        Yields
        ------
        list of IndOrT
            Path of oriented contigs
        """
        for region_path in self.all_region_paths():
            yield list(self.region_path_to_oriented_contigs(region_path))

    def region_path_to_oriented_contigs(
            self, region_path: Iterable[IndOrT]) -> Iterator[OrCT]:
        """Iterate over the oriented contigs of a given region path.

        Parameters
        ----------
        region_path : iterable of IndOrT
            Path of oriented regions

        Yields
        ------
        OrCT
            Oriented contig
        """
        for reg_ind, reg_or in region_path:
            yield from self.oriented_contigs_of_region(reg_ind, reg_or)

    # ~*~ Getter ~*~

    def revsymg(self) -> RevSymGraph:
        """Return the reverse symmetric graph associated.

        Returns
        -------
        RevSymGraph
            Reverse symmetric graph
        """
        return self.__graph

    def oriented_contigs_of_region(self, region_ind: IndexT,
                                   orientation: OrT = FORWARD_INT) -> (
            Iterator[OrCT]):
        """Iterate over oriented contigs of the oriented region.

        Parameters
        ----------
        region_ind : IndexT
            Region's index
        orientation : OrT, optional
            Region's orientation, by default FORWARD_INT

        Yields
        ------
        OrCT
            Oriented contig of the oriented region
        """
        if orientation == FORWARD_INT:
            yield from self.__regions_contigs[region_ind]
        else:
            for oriented_contig in reversed(self.__regions_contigs[region_ind]):
                yield rev_oriented_contig(oriented_contig)

    def region_path_length(self) -> int:
        """Return the length of all the region paths.

        Returns
        -------
        int
            _description_
        """
        return self.__region_path_length

    # ~*~ Private ~*~

    def __read_regions_contigs(self, contigs_of_regions_path: Path):
        """Extract the oriented contigs for each region.

        Parameters
        ----------
        contigs_of_regions_path : Path
            List of oriented contigs for each region
        """
        vertices = self.__graph.vertices()
        for region_ind, oriented_contigs in enumerate(
                read_contigs_of_regions(contigs_of_regions_path)):
            assert region_ind == vertices.add()
            self.__regions_contigs.append(oriented_contigs)

    def __add_regions_links(self, map_of_regions_path: Path):
        """Add regions links from map of regions.

        Parameters
        ----------
        map_of_regions_path : Path
            Map of regions
        """
        edges = self.__graph.edges()
        reg_indor_iter = read_map_of_regions(map_of_regions_path)
        start_indor = next(reg_indor_iter)
        u_indor = start_indor
        for v_indor in reg_indor_iter:
            edges.add(u_indor, v_indor)
            u_indor = v_indor
        edges.add(u_indor, start_indor)
        self.__region_path_length = len(self.__graph.edges()) // 2


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
def rev_oriented_contig(oriented_contig: OrCT) -> OrCT:
    """Return the reverse of the oriented contig.

    Parameters
    ----------
    oriented_contig : OrCT
        Oriented contig

    Returns
    -------
    OrCT
        Its reverse
    """
    return oriented_contig[ORC_ID_IND], ORIENT_REV[oriented_contig[ORC_OR_IND]]


# ---------------------------------------------------------------------------- #
#                               Region Paths File                              #
# ---------------------------------------------------------------------------- #
def write_region_paths(region_paths: Iterable[list[IndOrT]],
                       region_paths_file: Path):
    """Write the region paths file.

    Parameters
    ----------
    region_paths : iterable of list of IndOrT
        Region paths
    region_paths_file : Path
        Region paths file path
    """
    with open(region_paths_file, 'w', encoding='utf-8') as f_out:
        for region_path in region_paths:
            line: str = ''
            for reg_ind, reg_or in region_path:
                line += f'{reg_ind}\t{ORIENT_INT_STR[reg_or]}\t'
            f_out.write(line[:-1] + '\n')


def read_region_paths(region_paths_file: Path) -> Iterator[list[IndOrT]]:
    """Read region paths from file.

    Parameters
    ----------
    region_paths_file : Path
        Region paths file path

    Yields
    ------
    list of IndOrT
        Region path
    """
    with open(region_paths_file, 'r', encoding='utf-8') as regpaths_in:
        for line in regpaths_in:
            region_path = []
            l_regindor = line.split()
            k = 0
            while k < len(l_regindor) - 1:
                region_path.append(
                    (
                        IndexT(l_regindor[k]),
                        STR_ORIENT[l_regindor[k + 1]],  # type: ignore
                    ),
                )
                k += 2
            yield region_path


def fmt_region_paths_filename(instance_name: str,
                              ilp_combination: Iterable[str]) -> str:
    """Format the region paths filename.

    Parameters
    ----------
    instance_name : str
        Instance name
    ilp_combination : iterable of str
        ILP string code combination

    Returns
    -------
    str
        Formatted filename
    """
    return (
        f'{REGION_PATHS_PREFIX}_{instance_name}'
        f"_{'_'.join(ilp_combination)}"
        f'.{REGION_PATHS_EXT}'
    )


# ---------------------------------------------------------------------------- #
#                          Oriented Contig Paths File                          #
# ---------------------------------------------------------------------------- #
def write_oriented_contig_paths(oriented_contig_paths: Iterable[list[OrCT]],
                                oriented_contig_paths_file: Path):
    """Write the oriented contig paths file.

    Parameters
    ----------
    oriented_contig_paths : iterable of list of OrCT
        Region paths
    oriented_contig_paths_file : Path
        Region paths file path
    """
    with open(oriented_contig_paths_file, 'w', encoding='utf-8') as f_out:
        for oriented_contig_path in oriented_contig_paths:
            line: str = ''
            for c_id, c_or in oriented_contig_path:
                line += f'{c_id}\t{ORIENT_INT_STR[c_or]}\t'
            f_out.write(line[:-1] + '\n')


def read_oriented_contig_paths(oriented_contig_paths_file: Path) -> (
        Iterator[list[OrCT]]):
    """Read region paths from file.

    Parameters
    ----------
    oriented_contig_paths_file : Path
        Region paths file path

    Yields
    ------
    list of OrCT
        Oriented contig path
    """
    with open(oriented_contig_paths_file, 'r', encoding='utf-8') as regpaths_in:
        for line in regpaths_in:
            oriented_contig_path = []
            l_orc = line.split()
            k = 0
            while k < len(l_orc) - 1:
                oriented_contig_path.append(
                    (
                        l_orc[k],
                        STR_ORIENT[l_orc[k + 1]],  # type: ignore
                    ),
                )
                k += 2
            yield oriented_contig_path


def fmt_oriented_contig_paths_filename(instance_name: str,
                                       ilp_combination: Iterable[str]) -> str:
    """Format the oriented contig paths filename.

    Parameters
    ----------
    instance_name : str
        Instance name
    ilp_combination : iterable of str
        ILP string code combination

    Returns
    -------
    str
        Formatted filename
    """
    return (
        f'{ORIENTED_CONTIG_PATHS_PREFIX}_{instance_name}'
        f"_{'_'.join(ilp_combination)}"
        f'.{ORIENTED_CONTIG_PATHS_EXT}'
    )



# -*- coding=utf-8 -*-

"""Unit testing for assembly graph."""

# pylint: disable=compare-to-zero, missing-raises-doc
from pathlib import Path

from revsymg.graphs import RevSymGraph
from revsymg.index_lib import FORWARD_INT, REVERSE_INT, IndOrT

from khloraascaf.assembly_graph import (
    AssemblyGraph,
    OrCT,
    fmt_oriented_contig_paths_filename,
    fmt_region_paths_filename,
    read_oriented_contig_paths,
    read_region_paths,
    rev_oriented_contig,
    write_oriented_contig_paths,
    write_region_paths,
)


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
TEST_DIR: Path = Path(__file__).parent.absolute()

_TOY_DATADIR = TEST_DIR / 'data'

# ---------------------------------------------------------------------------- #
#                                    IR - UN                                   #
# ---------------------------------------------------------------------------- #
_IR_UN_DIR = _TOY_DATADIR / 'ir_un'
_IR_UN_SOL_REGMAP = _IR_UN_DIR / 'map_of_regions_sol.tsv'
_IR_UN_SOL_REGCTG_F = _IR_UN_DIR / 'contigs_of_regions_sol_0.tsv'
_IR_UN_REG_PATHS = _IR_UN_DIR / 'region_paths.tsv'
_IR_UN_ORC_PATHS = _IR_UN_DIR / 'oriented_contig_paths.tsv'

# ---------------------------------------------------------------------------- #
#                                    DR - UN                                   #
# ---------------------------------------------------------------------------- #
_DR_UN_DIR = _TOY_DATADIR / 'dr_un'
_DR_UN_SOL_REGMAP = _DR_UN_DIR / 'map_of_regions_sol.tsv'
_DR_UN_SOL_REGCTG = _DR_UN_DIR / 'contigs_of_regions_sol.tsv'
_DR_UN_REG_PATHS = _DR_UN_DIR / 'region_paths.tsv'
_DR_UN_ORC_PATHS = _DR_UN_DIR / 'oriented_contig_paths.tsv'


# ---------------------------------------------------------------------------- #
#                                      UN                                      #
# ---------------------------------------------------------------------------- #
_UN_DIR = _TOY_DATADIR / 'un'
_UN_SOL_REGMAP = _UN_DIR / 'map_of_regions_sol.tsv'
_UN_SOL_REGCTG = _UN_DIR / 'contigs_of_regions_sol.tsv'


# ============================================================================ #
#                                TEST FUNCTIONS                                #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                               All_region_paths                               #
# ---------------------------------------------------------------------------- #
def test_all_region_paths_ir_un():
    """Test all_region_paths method for IR-UN."""
    # TOTEST use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _IR_UN_SOL_REGMAP,
        _IR_UN_SOL_REGCTG_F,
    )
    s_region_paths: set[tuple[IndOrT, ...]] = {
        (
            (0, FORWARD_INT),
            (1, FORWARD_INT),
            (2, FORWARD_INT),
            (1, REVERSE_INT),
        ),
        (
            (0, FORWARD_INT),
            (1, FORWARD_INT),
            (2, REVERSE_INT),
            (1, REVERSE_INT),
        ),
    }
    asm_graph_paths = [tuple(l) for l in asm_graph.all_region_paths()]
    assert len(asm_graph_paths) == len(s_region_paths)
    for reg_path in asm_graph_paths:
        assert reg_path in s_region_paths


def test_all_region_paths_dr_un():
    """Test all_region_paths method for DR-UN."""
    # TOTEST use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _DR_UN_SOL_REGMAP,
        _DR_UN_SOL_REGCTG,
    )
    s_region_paths: set[tuple[IndOrT, ...]] = {
        (
            (0, FORWARD_INT),
            (1, FORWARD_INT),
            (2, FORWARD_INT),
            (1, FORWARD_INT),
        ),
    }
    asm_graph_paths = [tuple(l) for l in asm_graph.all_region_paths()]
    assert len(asm_graph_paths) == len(s_region_paths)
    for reg_path in asm_graph_paths:
        assert reg_path in s_region_paths


# TOTEST test_all_regions_paths_un
def test_all_region_paths_un():
    """Test all_region_paths method for UN."""
    # TOTEST use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _UN_SOL_REGMAP,
        _UN_SOL_REGCTG,
    )
    s_region_paths: set[tuple[IndOrT, ...]] = {
        (
            (0, FORWARD_INT),
        ),
    }
    asm_graph_paths = [tuple(l) for l in asm_graph.all_region_paths()]
    assert len(asm_graph_paths) == len(s_region_paths)
    for reg_path in asm_graph_paths:
        assert reg_path in s_region_paths


# ---------------------------------------------------------------------------- #
#                           All_oriented_contig_paths                          #
# ---------------------------------------------------------------------------- #
def test_all_oriented_contig_paths_ir_un():
    """Test all_oriented_contig_paths method for IR-UN."""
    # TOTEST use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _IR_UN_SOL_REGMAP,
        _IR_UN_SOL_REGCTG_F,
    )
    s_orc_paths: set[tuple[OrCT, ...]] = {
        (
            ('C0', FORWARD_INT),
            ('C1', REVERSE_INT),
            ('C2', REVERSE_INT),
            ('C3', FORWARD_INT),
            ('C4', REVERSE_INT),
            ('C3', REVERSE_INT),
            ('C2', FORWARD_INT),
        ),
        (
            ('C0', FORWARD_INT),
            ('C1', REVERSE_INT),
            ('C2', REVERSE_INT),
            ('C3', FORWARD_INT),
            ('C4', FORWARD_INT),
            ('C3', REVERSE_INT),
            ('C2', FORWARD_INT),
        ),
    }
    asm_graph_paths = [tuple(l) for l in asm_graph.all_oriented_contig_paths()]
    assert len(asm_graph_paths) == len(s_orc_paths)
    for reg_path in asm_graph_paths:
        assert reg_path in s_orc_paths


def test_all_oriented_contig_paths_dr():
    """Test all_oriented_contig_paths method for DR-UN."""
    # TOTEST use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _DR_UN_SOL_REGMAP,
        _DR_UN_SOL_REGCTG,
    )
    s_orc_paths: set[tuple[OrCT, ...]] = {
        (
            ('C0', FORWARD_INT),
            ('C1', REVERSE_INT),
            ('C2', REVERSE_INT),
            ('C3', FORWARD_INT),
            ('C4', FORWARD_INT),
            ('C2', REVERSE_INT),
            ('C3', FORWARD_INT),
        ),
    }
    asm_graph_paths = [tuple(l) for l in asm_graph.all_oriented_contig_paths()]
    assert len(asm_graph_paths) == len(s_orc_paths)
    for reg_path in asm_graph_paths:
        assert reg_path in s_orc_paths


# ---------------------------------------------------------------------------- #
#                                    Revsymg                                   #
# ---------------------------------------------------------------------------- #
def test_getter_revsymg():
    """Test getter revsymg."""
    # TOTEST use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _IR_UN_SOL_REGMAP,
        _IR_UN_SOL_REGCTG_F,
    )
    # TOTEST use fixtures and composition
    # TOTEST find a better assertion
    assert isinstance(asm_graph.revsymg(), RevSymGraph)


# ---------------------------------------------------------------------------- #
#                          Oriented_contigs_of_region                          #
# ---------------------------------------------------------------------------- #
def test_oriented_contigs_of_region():
    """Test oriented_contigs_of_region method."""
    # TOTEST use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _IR_UN_SOL_REGMAP,
        _IR_UN_SOL_REGCTG_F,
    )
    # TOTEST use fixtures and composition
    assert list(asm_graph.oriented_contigs_of_region(0)) == [
        ('C0', FORWARD_INT),
        ('C1', REVERSE_INT),
    ]
    assert list(asm_graph.oriented_contigs_of_region(0, FORWARD_INT)) == [
        ('C0', FORWARD_INT),
        ('C1', REVERSE_INT),
    ]
    assert list(asm_graph.oriented_contigs_of_region(0, REVERSE_INT)) == [
        ('C1', FORWARD_INT),
        ('C0', REVERSE_INT),
    ]


# ---------------------------------------------------------------------------- #
#                              Rev_oriented_contig                             #
# ---------------------------------------------------------------------------- #
def test_rev_oriented_contig():
    """Test rev_oriented_contig function."""
    assert rev_oriented_contig(('0', FORWARD_INT)) == ('0', REVERSE_INT)
    assert rev_oriented_contig(('3', REVERSE_INT)) == ('3', FORWARD_INT)


# ---------------------------------------------------------------------------- #
#                              Write_region_paths                              #
# ---------------------------------------------------------------------------- #
def test_write_region_paths():
    """Test write_region_paths function."""
    asm_graph = AssemblyGraph(
        _IR_UN_SOL_REGMAP,
        _IR_UN_SOL_REGCTG_F,
    )
    region_paths_file = Path('tmp_region_paths.tsv')
    write_region_paths(asm_graph.all_region_paths(), region_paths_file)
    with open(region_paths_file, 'r', encoding='utf-8') as f_in:
        to_test_lines: set[tuple[str, ...]] = {
            tuple(line.split()) for line in f_in
        }
    with open(_IR_UN_REG_PATHS, 'r', encoding='utf-8') as f_in:
        sol_lines: set[tuple[str, ...]] = {
            tuple(line.split()) for line in f_in
        }
    assert to_test_lines == sol_lines
    region_paths_file.unlink()


# ---------------------------------------------------------------------------- #
#                               Read_region_paths                              #
# ---------------------------------------------------------------------------- #
def test_read_region_paths():
    """Test read_region_paths function."""
    region_paths = read_region_paths(_IR_UN_REG_PATHS)
    s_region_paths: set[tuple[IndOrT, ...]] = {
        (
            (0, FORWARD_INT),
            (1, FORWARD_INT),
            (2, FORWARD_INT),
            (1, REVERSE_INT),
        ),
        (
            (0, FORWARD_INT),
            (1, FORWARD_INT),
            (2, REVERSE_INT),
            (1, REVERSE_INT),
        ),
    }
    for region_path in region_paths:
        assert tuple(region_path) in s_region_paths


# ---------------------------------------------------------------------------- #
#                           Fmt_region_paths_filename                          #
# ---------------------------------------------------------------------------- #
def test_fmt_region_paths_file():
    """Test fmt_region_paths_filename."""
    assert fmt_region_paths_filename('jaaj', ('ir', 'un')) == (
        'region_paths_jaaj_ir_un.tsv'
    )


# ---------------------------------------------------------------------------- #
#                          Write_oriented_contig_paths                         #
# ---------------------------------------------------------------------------- #
def test_write_oriented_contig_paths():
    """Test write_oriented_contig_paths function."""
    asm_graph = AssemblyGraph(
        _IR_UN_SOL_REGMAP,
        _IR_UN_SOL_REGCTG_F,
    )
    oriented_contig_paths_file = Path('tmp_oriented_contig_paths.tsv')
    write_oriented_contig_paths(
        asm_graph.all_oriented_contig_paths(),
        oriented_contig_paths_file)
    with open(oriented_contig_paths_file, 'r', encoding='utf-8') as f_in:
        to_test_lines: set[tuple[str, ...]] = {
            tuple(line.split()) for line in f_in
        }
    with open(_IR_UN_ORC_PATHS, 'r', encoding='utf-8') as f_in:
        sol_lines: set[tuple[str, ...]] = {
            tuple(line.split()) for line in f_in
        }
    assert to_test_lines == sol_lines
    oriented_contig_paths_file.unlink()


# ---------------------------------------------------------------------------- #
#                          Read_oriented_contig_paths                          #
# ---------------------------------------------------------------------------- #
def test_read_oriented_contig_paths():
    """Test read_oriented_contig_paths function."""
    oriented_contig_paths = read_oriented_contig_paths(_IR_UN_ORC_PATHS)
    s_oriented_contig_paths: set[tuple[OrCT, ...]] = {
        (
            ('C0', FORWARD_INT),
            ('C1', REVERSE_INT),
            ('C2', REVERSE_INT),
            ('C3', FORWARD_INT),
            ('C4', REVERSE_INT),
            ('C3', REVERSE_INT),
            ('C2', FORWARD_INT),
        ),
        (
            ('C0', FORWARD_INT),
            ('C1', REVERSE_INT),
            ('C2', REVERSE_INT),
            ('C3', FORWARD_INT),
            ('C4', FORWARD_INT),
            ('C3', REVERSE_INT),
            ('C2', FORWARD_INT),
        ),
    }
    for oriented_contig_path in oriented_contig_paths:
        assert tuple(oriented_contig_path) in s_oriented_contig_paths


# ---------------------------------------------------------------------------- #
#                      Fmt_oriented_contig_paths_filename                      #
# ---------------------------------------------------------------------------- #
def test_fmt_oriented_contig_paths_file():
    """Test fmt_oriented_contig_paths_filename."""
    assert fmt_oriented_contig_paths_filename('jaaj', ('ir', 'un')) == (
        'oriented_contig_paths_jaaj_ir_un.tsv'
    )



# -*- coding=utf-8 -*-

"""PuLP circuit model."""


from pulp import LpProblem, lpSum
from revsymg.index_lib import FORWARD_INT, REVERSE_INT

from khloraascaf.ilp.pulp_var_db import PuLPVarPath
from khloraascaf.multiplied_doubled_contig_graph import (
    CIND_IND,
    MULT_ATTR,
    MDCGraph,
    OccOrCT,
    rev_occorc,
)


# ============================================================================ #
#                                  CONSTRAINTS                                 #
# ============================================================================ #
def flow_definition(prob: LpProblem, var: PuLPVarPath,
                    big_m: int, mdcg: MDCGraph):
    """Constraint defining the flow on the edges.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarPath
        Variables
    big_m : int
        Big M
    mdcg : MDCGraph
        Mulitplied doubled contig graph
    """
    for e in mdcg.multiplied_edges():
        #
        # If edge is chosen, then its flow is positive
        #
        prob += var.x[e] <= var.f[e]
        #
        # If edge is not chosen, then its flow is 0-negative
        #
        prob += var.f[e] <= big_m * var.x[e]


def circuit_from_the_starter_forward(prob: LpProblem, var: PuLPVarPath,
                                     mdcg: MDCGraph, starter_vertex: OccOrCT):
    """Constraint defining the path from start to terminal.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarPath
        Variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    """
    # ------------------------------------------------------------------------ #
    # Use the starter in forward orientation as the beginning of the circuit
    # ------------------------------------------------------------------------ #
    #
    # The path begins from the starter
    #
    prob += (
        lpSum(
            var.x[starter_vertex, v]
            for v in mdcg.multiplied_succs(starter_vertex)
        ) == 1
    )
    #
    # The path ends to the starter
    #
    prob += (
        lpSum(
            var.x[u, starter_vertex]
            for u in mdcg.multiplied_preds(starter_vertex)
        ) == 1
    )
    #
    # Flow is initialised to one
    #
    prob += (
        lpSum(
            var.f[starter_vertex, v]
            for v in mdcg.multiplied_succs(starter_vertex)
        ) == 1
    )
    # ------------------------------------------------------------------------ #
    # Avoid use of start reverse
    # ------------------------------------------------------------------------ #
    start_r = rev_occorc(starter_vertex)

    for u in mdcg.multiplied_preds(start_r):
        prob += var.x[u, start_r] == 0  # pylint: disable=compare-to-zero
    for v in mdcg.multiplied_succs(start_r):
        prob += var.x[start_r, v] == 0  # pylint: disable=compare-to-zero
    #
    # Flow is initialised to zero
    #
    prob += (
        lpSum(  # pylint: disable=compare-to-zero
            var.f[start_r, v]
            for v in mdcg.multiplied_succs(start_r)
        ) == 0
    )


def intermediate_in_circuit(prob: LpProblem, var: PuLPVarPath,
                            mdcg: MDCGraph, starter_vertex: OccOrCT):
    """Constraint defining the intermediate vertices in the circuit.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarPath
        Variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    """
    vertices = mdcg.vertices()

    # Intermediate are non-starter vertex (nor its reverse)
    for v_ind in (i for i in range(vertices.card_index())
                  if i != starter_vertex[CIND_IND]):

        for v_occ in range(vertices.attr(v_ind, MULT_ATTR)):

            v_f: OccOrCT = (v_ind, FORWARD_INT, v_occ)
            v_r: OccOrCT = (v_ind, REVERSE_INT, v_occ)
            #
            # Forward or reverse or nothing
            #
            prob += var.i[v_f] + var.i[v_r] <= 1

            for v in (v_f, v_r):
                #
                # Intermediate iif there is exactly one in-edge
                #   and one out-edge
                #
                prob += (
                    var.i[v]
                    == lpSum(var.x[u, v] for u in mdcg.multiplied_preds(v))
                )
                prob += (
                    var.i[v]
                    == lpSum(var.x[v, w] for w in mdcg.multiplied_succs(v))
                )
                #
                # Passing throw an intermediate increments its out-flow
                #
                prob += (
                    lpSum(var.f[v, w] for w in mdcg.multiplied_succs(v))
                    - lpSum(var.f[u, v] for u in mdcg.multiplied_preds(v))
                    == var.i[v]
                )
